package cyberdeliaengine

import android.content.res.Resources
import java.nio.ByteBuffer
import java.nio.ByteOrder
import java.nio.FloatBuffer
import android.opengl.Matrix
import android.opengl.GLES32
import android.opengl.GLUtils
import android.graphics.BitmapFactory
import android.graphics.Bitmap
import java.nio.IntBuffer
import android.util.Log



open class Model3D {

    val mVertexShaderCode = (
                    "uniform mat4 u_MVPMatrix;" +
                    "attribute vec4 a_Position;" +
                    "attribute vec2 a_TextureCoord;" +
                    "varying vec2 v_TextureCoord;" +
                    "attribute float a_Bone;" +
                    "uniform mat4 u_BoneMatrix[36];" +
                    "void main() {" +
                    "   int bone = int(a_Bone);" +
                    "   if ( bone >= 0 ) { " +
                    "       gl_Position = u_MVPMatrix * u_BoneMatrix[bone] * a_Position;" +
                    "   } else {" +
                    "       gl_Position = u_MVPMatrix * a_Position;" +
                    "   }" +
                    "   v_TextureCoord = a_TextureCoord;" +
                    "}")

    val mFragmentShaderCode = (
                    "precision mediump float;" +
                    "varying vec2 v_TextureCoord;" +
                    "uniform sampler2D u_Texture;" +
                    "void main() {" +
                    "  gl_FragColor = texture2D(u_Texture, v_TextureCoord);" +
                    "}")

    val mFloatSizeBytes: Int = 4
    val mBoneStrideBytes: Int = mFloatSizeBytes
    var mBoneBuffer:FloatBuffer? = null
    var mBoneHandle:Int = 0
    var uBoneMatrixHandle:Int = 0
    var aTextureHandle: Int = 0
    val mUvDataStrideBytes: Int = 2 * mFloatSizeBytes
    var mTextureBuffer: FloatBuffer? = null
    var mTextureID: Int = 0

    var mVertexBuffer: FloatBuffer? = null
    var mProgram: Int = 0
    var mPositionHandle: Int = 0
    var mMVPMatrixHandle: Int = 0
    var mUseBoneHandle: Int = 0

    val mIndices = ArrayList<Short>(0)
    val mCoords = ArrayList<Float>(0)
    val mUVs = ArrayList<Float>(0)
    var mCurrentAnimation: Int = 0

    val mCoordsPerVertex = 3
    val mVertexStrideBytes = mCoordsPerVertex * mFloatSizeBytes

    val mMVMatrix = FloatArray(16)
    val mMVPMatrix = FloatArray(16)
    private val mTranslateMatrix = FloatArray(16)
    private val mRotationMatrix = FloatArray(16)
    var mAngle: Float = 0.toFloat()
    var mTranslation = Vector3D(0f,0f,0f)
    var mDistance = 0f
    private var mDistance2 = 0f
    private var mIntersectU = 0f
    private var mIntersectV = 0f
    val mBones = ArrayList<Float>(0)

    init {
    }


    fun addV(x: Double, y: Double, z: Double) {
        mCoords.add(x.toFloat())
        mCoords.add(y.toFloat())
        mCoords.add(z.toFloat())
        mBones.add(-1f)
    }

    fun addI(i0: Short, i1: Short, i2: Short) {
        mIndices.add(i0)
        mIndices.add(i1)
        mIndices.add(i2)
    }

    fun addUV(u: Double, v: Double) {
        mUVs.add(u.toFloat())
        mUVs.add(v.toFloat())
    }

    open fun newModel3D() {
        val coords = FloatArray(mIndices.size * 3)
        for (i in mIndices.indices) {
            val v = mIndices[i] * 3
            coords[i * 3] = mCoords[v]
            coords[i * 3 + 1] = mCoords[v + 1]
            coords[i * 3 + 2] =  mCoords[v + 2]
        }
        val indices = ShortArray(mIndices.size)
        for (i in mIndices.indices) {
            indices[i] = i.toShort()
        }
        val uvs = FloatArray(mUVs.size)
        for (i in mUVs.indices) {
            uvs[i] = mUVs[i]
        }
        val bones = FloatArray(mIndices.size)
        for (i in mIndices.indices) {
            bones[i] = mBones[mIndices[i].toInt()]
        }

        val bb = ByteBuffer.allocateDirect(coords.size * mFloatSizeBytes)
        bb.order(ByteOrder.nativeOrder())
        mVertexBuffer = bb.asFloatBuffer()
        mVertexBuffer!!.put(coords)
        mVertexBuffer!!.position(0)

        val bbb = ByteBuffer.allocateDirect(bones.size * mFloatSizeBytes)
        bbb.order(ByteOrder.nativeOrder())
        mBoneBuffer = bbb.asFloatBuffer()
        mBoneBuffer!!.put(bones)
        mBoneBuffer!!.position(0)

        val uvbb = ByteBuffer.allocateDirect(
                mUVs.size * mFloatSizeBytes)
        uvbb.order(ByteOrder.nativeOrder())
        mTextureBuffer = uvbb.asFloatBuffer()
        for (i in mUVs.indices) {
            mTextureBuffer!!.put(mUVs[i])
        }
        mTextureBuffer!!.position(0)

        val vertexShader = loadShader(
                GLES32.GL_VERTEX_SHADER,
                mVertexShaderCode)
        val fragmentShader = loadShader(
                GLES32.GL_FRAGMENT_SHADER,
                mFragmentShaderCode)

        mProgram = GLES32.glCreateProgram()             // create empty OpenGL Program
        GLES32.glAttachShader(mProgram, vertexShader)   // add the vertex shader to program
        GLES32.glAttachShader(mProgram, fragmentShader) // add the fragment shader to program
        GLES32.glLinkProgram(mProgram)                  // create OpenGL program executables
    }

    open fun draw() {

        GLES32.glUseProgram(mProgram)

        mPositionHandle = GLES32.glGetAttribLocation(mProgram, "a_Position")

        GLES32.glEnableVertexAttribArray(mPositionHandle)

        GLES32.glVertexAttribPointer(
                mPositionHandle, mCoordsPerVertex,
                GLES32.GL_FLOAT, false,
                mVertexStrideBytes, mVertexBuffer)

        GLES32.glActiveTexture(GLES32.GL_TEXTURE0)
        GLES32.glBindTexture(GLES32.GL_TEXTURE_2D, mTextureID)

        mBoneHandle = GLES32.glGetAttribLocation(mProgram, "a_Bone")
        GLES32.glEnableVertexAttribArray(mBoneHandle)
        GLES32.glVertexAttribPointer(
            mBoneHandle, mCoordsPerVertex,
            GLES32.GL_FLOAT, false,
            mBoneStrideBytes, mBoneBuffer)

        aTextureHandle = GLES32.glGetAttribLocation(mProgram, "a_TextureCoord")
        GLES32.glEnableVertexAttribArray(aTextureHandle)
        GLES32.glVertexAttribPointer(
                aTextureHandle, 2,
                GLES32.GL_FLOAT, false,
                mUvDataStrideBytes, mTextureBuffer)

        mMVPMatrixHandle = GLES32.glGetUniformLocation(mProgram, "u_MVPMatrix")
        GLES32.glUniformMatrix4fv(mMVPMatrixHandle, 1, false, mMVPMatrix, 0)

        GLES32.glDrawArrays(GLES32.GL_TRIANGLES, 0, mIndices.size)

        GLES32.glDisableVertexAttribArray(mPositionHandle)
    }

    fun setTexture(res: Resources, id: Int) {
        val textures = IntArray(1)
        val `is` = res.openRawResource(id)

        val bitmap: Bitmap
        try {
            bitmap = BitmapFactory.decodeStream(`is`)
        } finally {
            try {
                `is`.close()
            } catch (e: Exception) {
            }

        }
        GLES32.glGenTextures(1, textures, 0)

        mTextureID = textures[0]
        GLES32.glBindTexture(GLES32.GL_TEXTURE_2D, mTextureID)

        GLES32.glTexParameterf(GLES32.GL_TEXTURE_2D,
                GLES32.GL_TEXTURE_MIN_FILTER, GLES32.GL_NEAREST.toFloat())
        GLES32.glTexParameterf(GLES32.GL_TEXTURE_2D,
                GLES32.GL_TEXTURE_MAG_FILTER, GLES32.GL_NEAREST.toFloat())

        GLES32.glTexParameteri(GLES32.GL_TEXTURE_2D,
                GLES32.GL_TEXTURE_WRAP_S, GLES32.GL_REPEAT)
        GLES32.glTexParameteri(GLES32.GL_TEXTURE_2D,
                GLES32.GL_TEXTURE_WRAP_T, GLES32.GL_REPEAT)

        GLUtils.texImage2D(GLES32.GL_TEXTURE_2D, 0, bitmap, 0)
    }

    fun transform(mViewMatrix:FloatArray,mProjectionMatrix:FloatArray) {
        Matrix.setIdentityM(mTranslateMatrix,0)
        Matrix.translateM(mTranslateMatrix, 0, mTranslation.x,mTranslation.y,mTranslation.z)
        Matrix.setRotateM(mRotationMatrix, 0, mAngle, 0f, 1.0f, 0f)
        Matrix.multiplyMM(mMVMatrix, 0, mTranslateMatrix, 0, mRotationMatrix, 0)
        Matrix.multiplyMM(mMVMatrix, 0, mViewMatrix, 0, mMVMatrix, 0)
        Matrix.multiplyMM(mMVPMatrix, 0, mProjectionMatrix, 0, mMVMatrix, 0)
    }

    fun loadShader(type: Int, shaderCode: String): Int {
       val shader = GLES32.glCreateShader(type)

        GLES32.glShaderSource(shader, shaderCode)
        GLES32.glCompileShader(shader)

        return shader
    }

    fun getPixels(x:Int,y:Int,width:Int,height:Int): IntArray {
        val pixels = IntArray(width * height)
        val buffer = IntBuffer.wrap(pixels)
        buffer.position(0)
        GLES32.glReadPixels(x,y,width,height, GLES32.GL_RGBA, GLES32.GL_UNSIGNED_BYTE, buffer)
        return pixels
    }

    fun intersect(origin: Vector3D, dir: Vector3D): Boolean {
        var flg = false
        mDistance = 1000000f
        for (i in mIndices.indices step 3) {
            val i0 = mIndices[i  ] * 3
            val i1 = mIndices[i+1] * 3
            val i2 = mIndices[i+2] * 3
            val v0 = Vector3D(mCoords[i0],mCoords[i0 + 1],mCoords[i0 + 2])
            val v1 = Vector3D(mCoords[i1],mCoords[i1 + 1],mCoords[i1 + 2])
            val v2 = Vector3D(mCoords[i2],mCoords[i2 + 1],mCoords[i2 + 2])
           if (intersectTriangle(origin, dir, v0, v1, v2)) {
               if ( mDistance2 >= 0 && mDistance > mDistance2 )
               {
                   mDistance = mDistance2
                   flg = true
               }
           }
        }

        return flg
    }

    private fun intersectTriangle(
            origin: Vector3D, dir: Vector3D,
            v0: Vector3D, v1: Vector3D, v2: Vector3D): Boolean {
        val e1 = v1 - v0
        val e2 = v2 - v0
        val pv: Vector3D
        val tv: Vector3D
        val qv: Vector3D
        val det: Float
        var t: Float
        var u: Float
        var v: Float
        val invDet: Float

        pv = dir.crossProduct(e2)
        det = e1.dotProduct(pv)

        if (det > 1e-3) {
            tv = origin - v0
            u = tv.dotProduct(pv)
            if (u < 0.0f || u > det) return false

            qv = tv.crossProduct(e1)

            v = dir.dotProduct(qv)
            if (v < 0.0 || u + v > det) return false
        } else if (det < -1e-3) {
            tv = origin - v0

            u = tv.dotProduct(pv)
            if (u > 0.0 || u < det) return false

            qv = tv.crossProduct(e1)

            v = dir.dotProduct(qv)
            if (v > 0.0 || u + v < det) return false

        } else {
            return false
        }

        invDet = 1.0f / det

        t = e2.dotProduct(qv)
        t *= invDet
        u *= invDet
        v *= invDet

        mDistance2 = t
        mIntersectU = u
        mIntersectV = v

        return true    //hit!!
    }

	fun mousePick(x:Float,y:Float,width:Int,height:Int,projection:FloatArray):Boolean
	{
		var vPickRayDir = Vector3D()
		var vPickRayOrig = Vector3D()
		var v = Vector3D()

		v.x =  ( ( ( 2 * x ) / width  ) - 1 ) / projection[0]
		v.y = -( ( ( 2 * y ) / height ) - 1 ) / projection[5]
		v.z = -1.0f

		var m:Matrix3D = Matrix3D()
		m = m.inverse(mMVMatrix)

		vPickRayDir.x  = v.x*m._e[0] + v.y*m._e[4] + v.z*m._e[8]
		vPickRayDir.y  = v.x*m._e[1] + v.y*m._e[5] + v.z*m._e[9]
		vPickRayDir.z  = v.x*m._e[2] + v.y*m._e[6] + v.z*m._e[10]
		vPickRayOrig.x = m._e[12]
		vPickRayOrig.y = m._e[13]
		vPickRayOrig.z = m._e[14]

		if( intersect(vPickRayOrig,vPickRayDir) )
		{
			return true
		}

		return false
	}
}
